from coalib.parsing.filters.Filter import filter


@filter
def language_filter(bear, args):
    """
    Filters the bears by ``LANGUAGES``.

    :param bear: Bear object.
    :param args: Set of languages on which ``bear`` is to be filtered.
    :return:     ``True`` if this bear matches the criteria inside args,
                 ``False`` otherwise.
    """
    return bool({lang.lower() for lang in bear.LANGUAGES} & (args | {'all'}))



from coalib.results.result_actions.ResultAction import ResultAction
from coalib.results.result_actions.ShowPatchAction import ShowPatchAction
from coalib.results.RESULT_SEVERITY import (
    RESULT_SEVERITY, RESULT_SEVERITY_COLORS)
from pyprint.ConsolePrinter import ConsolePrinter


def format_lines(lines, symbol='', line_nr=''):
    def sym(x): return ']' if x is '[' else x
    return '\n'.join('{}{:>5}{} {}'.format(symbol, sym(symbol), line_nr, line)
                     for line in lines.rstrip('\n').split('\n'))


class ShowAppliedPatchesAction(ResultAction):

    SUCCESS_MESSAGE = 'All actions have been applied'

    is_applicable = staticmethod(ShowPatchAction.is_applicable)

    def apply(self,
              result,
              original_file_dict,
              file_diff_dict):
        """
        Show Applied (P)atches
        """
        console_printer = ConsolePrinter()
        applied_actions = result.get_applied_actions()
        show_patch_action = ShowPatchAction()
        RESULT_INDEX = 0
        FILE_DICT_INDEX = 1
        FILE_DIFF_DICT_INDEX = 2
        SECTION_INDEX = 3

        for key, val in applied_actions.items():
            this_result = val[RESULT_INDEX]
            this_section = val[SECTION_INDEX]
            color_res = RESULT_SEVERITY_COLORS[this_result.severity]
            console_printer.print('\n**** {bear} [Section: {section}] ***'
                                  '*\n**** Action Applied: {action} ****\n'
                                  .format(bear=this_result.origin,
                                          section=this_section.name,
                                          action=key),
                                  color=color_res)
            console_printer.print(format_lines('[Severity: {sev}]'.format(
                sev=RESULT_SEVERITY.__str__(this_result.severity)), '!'),
                  color=color_res)
            show_patch_action.apply_from_section(val[RESULT_INDEX],
                                                 val[FILE_DICT_INDEX],
                                                 val[FILE_DIFF_DICT_INDEX],
                                                 val[SECTION_INDEX])
            console_printer.print(
                '\n**************\n', color=color_res)
        return True



from coalib.bearlib.aspects import Root, Taste


@Root.subaspect
class Formatting:
    """
    The visual appearance of source code.
    """
    class docs:
        example = """
        # Here is an example of Python code with lots of
        # formatting issues including: trailing spaces, missing spaces
        # around operators, strange and inconsistent indentation etc.

        z = 'hello'+'world'
             def f ( a):
                pass
        """
        example_language = 'Python'
        importance_reason = """
        A coding style (the of rules or guidelines used when writing the
        source code) can drastically affect the readability, and
        maintainability of a program and might as well introduce bugs.
        """
        fix_suggestions = """
        Defining a clearly and thoughtful coding style (based on the available
        ones given the programming language in use) and strictly respect it or
        apply it through out the implementation of a project.
        """


@Formatting.subaspect
class Length:
    """
    Hold sub-aspects for file and line length.
    """
    class docs:
        example = """
        # We assume that the maximum number of characters per line is 10
        # and that the maximum number of lines per files is 3.

        def run(bear, file, filename, aspectlist):
            return bear.run(file, filename, aspectlist)
        """
        example_language = 'Python'
        importance_reason = """
        Too long lines of code and too large files result in code difficult to
        read, understand and maintain.
        """
        fix_suggestions = """
        Length issues can be fixed by writing shorter lines of code (splitting
        long lines into multiple shorter lines); writing shorter files
        (splitting files into modules, writing shorter methods and classes.).
        """


@Length.subaspect
class LineLength:
    """
    Number of characters found in a line of code.
    """
    class docs:
        example = """
        print('The length of this line is 38')
        """
        example_langague = 'Python'
        importance_reason = """
        Too long lines make code very difficult to read and maintain.
        """
        fix_suggestions = """
        Splitting long lines of code into multiple shorter lines whenever
        possible. Avoiding the usage of in-line language specific constructs
        whenever they result in too long lines.
        """
    max_line_length = Taste[int](
        'Maximum number of character for a line.',
        (79, 80, 100, 120, 160), default=80)


@Length.subaspect
class FileLength:
    """
    Number of lines found in a file.
    """
    class docs:
        example = """
        # This file would be a large file if we assume that the max number of
        # lines per file is 10

        class Node:
            def __init__(self, value, left_most_child, left_sibling):
                self.value=value
                self.left_most_child=left_most_child
                self.left_sibling=left_sibling

        # This is example is just showing what this aspect is about, because
        # the max number of lines per file is usually 999.
        """
        example_language = 'Python 3'
        importance_reason = """
        Too long programs (or files) are difficult to read, maintain and
        understand.
        """
        fix_suggestions = """
        Splitting files into modules, writing shorter methods and classes.
        """
    max_file_length = Taste[int](
        'Maximum number of line for a file',
        (999,), default=999)


@Formatting.subaspect
class Spacing:
    """
    All whitespace found between non-whitespace characters.
    """
    class docs:
        example = """
        # Here is an example of code with spacing issues including
        # unnecessary blank lines and missing space around operators.



        def func(   ):
           return      37*-+2
        """
        example_language = 'Python'
        importance_reason = """
        Useless spacing affects the readability and maintainability of a code.
        """
        fix_suggestions = """
        Removing the trailing spaces and the meaningless blank lines.
        """


@Spacing.subaspect
class TrailingSpace:
    """
    Unnecessary whitespace at end of a line.

    Trailing space is all whitespace found after the last non-whitespace
    character on the line until the newline. This includes tabs "\\\\t",
    blank lines, blanks etc.
    """
    class docs:
        example = """
        def func( a ):
              pass

        """.replace('\n', '\t\n')
        example_language = 'Python'
        importance_reason = """
        Trailing spaces make code less readable and maintainable.
        """
        fix_suggestions = """
        Removing the trailing spaces.
        """
    allow_trailing_spaces = Taste[bool](
        'Determines whether or not trailing spaces should be allowed or not.',
        (True, False), default=False)


@Spacing.subaspect
class BlankLine:
    """
    A line with zero characters.
    """
    class docs:
        example = """
        name = input('What is your name?')


        print('Hi, {}'.format(name))
        """
        example_language = 'Python 3'
        importance_reason = """
        Various programming styles use blank lines in different places.
        The usage of blank lines affects the readability, maintainability and
        length of a code i.e blank lines can either make code longer, less
        readable and maintainable or do the reverse.
        """
        fix_suggestions = """
        Following specific rules about the usage of blank lines: using them
        only when necessary.
        """


@BlankLine.subaspect
class BlankLineAfterDeclaration:
    """
    Those found after declarations.
    """
    class docs:
        example = """
        #include <stdio.h>

        int main ()
        {
          int a;
          float b;

          scanf("%d%f", &a, &b);
          printf("a = %d and b = %f", a, b);
          return 0;
        }
        """
        example_language = 'C'
        importance_reason = """
        Having a specific and reasonable number of blank lines after every
        block of declarations improves on the readability of the code.
        """
        fix_suggestions = """
        `BlankLintAfterDeclaration` issues can be fixed specifying (and of
        course using) a reasonable number of blank lines to use after block
        declaration.
        """
    blank_lines_after_declarations = Taste[int](
        'Represents the number of blank lines after declarations',
        (0, 1, 2), default=0)


@BlankLine.subaspect
class BlankLineAfterProcedure:
    """
    Those found after procedures or functions.
    """
    class docs:
        example = """
        #include <stdio.h>

        void proc(void){
            printf("this does nothing");
        } int add(float a, float b){
            return a + b;
        }
        """
        example_language = 'C'
        importance_reason = """
        Having a specific and reasonable number of blank lines after every
        procedures improves on the readability of the code.
        """
        fix_suggestions = """
        `BlankLintAfterProcedure` issues can be fixed specifying (and of
        course using) a reasonable number of blank lines to use after
        procedures' definition.
        """
    blank_lines_after_procedures = Taste[int](
        'Represents the number of blank lines to use after a procedure or'
        'a function', (0, 1, 2), default=1)


@BlankLine.subaspect
class BlankLineAfterClass:
    """
    Those found after classes' definitions.
    """
    class docs:
        example = """
        class SomeClass:
            def __init__(self):
                raise RuntimeError('Never instantiate this class')


        def func():
            pass
        """
        example_language = 'Python 3'
        importance_reason = """
        Having a specific number of blank lines after every classes'
        definitions declarations improves on the readability of the code.
        """
        fix_suggestions = """
        """
    blank_lines_after_class = Taste[int](
        'Represents the number of blank lines to use after a class'
        'definition.', (1, 2), default=2)


@BlankLine.subaspect
class NewlineAtEOF:
    """
    Newline character (usually '\\\\n', aka CR) found at the end of file.
    """
    class docs:
        example = """
        def do_nothing():
            pass
        """ + ('\n')
        example_language = 'Python'
        importance_reason = """
        A text file consists of a series of lines, each of which ends with a
        newline character (\\\\n). A file that is not empty and does not end
        with a newline is therefore not a text file.

        It's not just bad style, it can lead to unexpected behavior, utilities
        that are supposed to operate on text files may not cope well with files
        that don't end with a newline.
        """
        fix_suggestions = """
        `NewlineAtEOF` issues can be fixed by simply adding a newline at the
        end of the file.
        """
    newline_at_EOF = Taste[bool](
        'If ``True``, enforce a newline at End Of File.',
        (True, False), default=True)


@Spacing.subaspect
class SpacesAroundOperator:
    """
    Spacing around operators.
    """
    class docs:
        example = """
        def f(a, x):
            return 37+a[42 -  x]
        """
        example_language = 'Python'
        importance_reason = """
        Having a specific and reasonable number of whitespace (blank) around
        operators improves on the readability of the code.
        """
        fix_suggestions = """
        `SpacesAroundOperator` issues can be fixed by simply specifying and
        the number of whitespace to be used after each operator.
        """
    spaces_around_operators = Taste[int](
        'Represents the number of space to be used around operators.',
        (0, 1), default=1)
    spaces_before_colon = Taste[int](
        'Represents the number of blank spaces before colons.',
        (0, 1), default=0)
    spaces_after_colon = Taste[int](
        'Represents the number of blank spaces after colons.',
        (0, 1), default=1)


@Formatting.subaspect
class Quotation:
    """
    Quotation mark used for strings and docstrings.
    """
    class docs:
        example = """
        # Here is an example of code where both '' and "" quotation mark
        # Are used.

        string = 'coala is always written with lowercase c.'
        string = "coala is always written with lowercase c."
        """
        example_language = 'Python'
        importance_reason = """
        Using the same quotation whenever possible in the code, improve on its
        readability by introducing consistency.
        """
        fix_suggestions = """
        Choosing a preferred quotation and using it everywhere (if possible).
        """
    preferred_quotation = Taste[str](
        'Represents the preferred quotation',
        ('\'', '"'), default='\'')



from coalib.results.Result import Result
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.bearlib.aspects import aspectbase
from coala_utils.decorators import (
    enforce_signature, generate_ordering, generate_repr)


@generate_repr(('id', hex),
               'origin',
               'message',
               'contents')
@generate_ordering('contents',
                   'origin',
                   'message_base')
class HiddenResult(Result):
    """
    This is a result that is not meant to be shown to the user. It can be used
    to transfer any data from a dependent bear to others.
    """

    @enforce_signature
    def __init__(self,
                 origin,
                 contents,
                 message: str='',
                 affected_code: (tuple, list)=(),
                 severity: int=RESULT_SEVERITY.NORMAL,
                 additional_info: str='',
                 debug_msg='',
                 diffs: (dict, None)=None,
                 confidence: int=100,
                 aspect: (aspectbase, None)=None,
                 message_arguments: dict={},
                 applied_actions: dict={}):
        """
        Creates a new HiddenResult. The contents can be accessed with
        obj.contents later.

        :param origin:
            The originating bear.
        :param contents:
            Any object to send additional data (arbitrary python objects)
            besides a message to the dependent bear. The data has to
            be picklable.
        :param affected_code:
            A tuple of ``SourceRange`` objects pointing to related positions in
            the source code.
        :param severity:
            Severity of this result.
        :param additional_info:
            A long description holding additional information about the issue
            and/or how to fix it. You can use this like a manual entry for a
            category of issues.
        :param debug_msg:
            A message which may help the user find out why this result was
            yielded.
        :param diffs:
            A dictionary with filename as key and ``Diff`` object
            associated with it as value.
        :param confidence:
            A number between 0 and 100 describing the likelihood of this result
            being a real issue.
        :param aspect:
            An aspectclass instance which this result is associated to.
            Note that this should be a leaf of the aspect tree!
            (If you have a node, spend some time figuring out which of
            the leafs exactly your result belongs to.)
        :param message_arguments:
            Arguments to be provided to the base message.
        :param applied_actions:
            A dictionary that contains the result, file_dict, file_diff_dict and
            the section for an action.
        """
        Result.__init__(self,
                        origin,
                        message,
                        affected_code,
                        severity,
                        additional_info,
                        debug_msg,
                        diffs,
                        confidence,
                        aspect,
                        message_arguments,
                        applied_actions)

        self.contents = contents



from collections import defaultdict

from coalib.bearlib.aspects.collections import AspectList
from coalib.bearlib.languages.Language import Languages


class bearclass(type):
    """
    Metaclass for :class:`coalib.bears.Bear.Bear` and therefore all bear
    classes.

    Pushing bears into the future... ;)
    """

    # by default a bear class has no aspects
    aspects = defaultdict(lambda: AspectList([]))

    def __new__(mcs, clsname, bases, clsattrs, *varargs, aspects=None,
                languages=None):
        return type.__new__(mcs, clsname, bases, clsattrs, *varargs)

    def __init__(cls, clsname, bases, clsattrs, *varargs, aspects=None,
                 languages=None):
        """
        Initializes the ``.aspects`` dict and ``.languages`` array on new
        bear classes from the mapping and the sequence given to the
        keyword-only `aspects` and `languages` arguments, respectively.
        """
        type.__init__(cls, clsname, bases, clsattrs, *varargs)
        if aspects is not None:
            cls.languages = Languages(languages)
            cls.aspects = defaultdict(
                lambda: AspectList([]),
                ((k, AspectList(v, languages=cls.languages))
                 for (k, v) in dict(aspects).items()))



# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License
# for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import sys

from pyprint.ConsolePrinter import ConsolePrinter

from dependency_management.requirements.PipRequirement import PipRequirement

from coalib.parsing.FilterHelper import (
    apply_filter, apply_filters, InvalidFilterException)
from coalib.output.Logging import configure_logging
from coalib.parsing.DefaultArgParser import default_arg_parser
from coalib.misc.Exceptions import get_exitcode


def main(debug=False):
    configure_logging()

    args = None  # to have args variable in except block when parse_args fails
    try:
        # Note: We parse the args here once to check whether to show bears or
        # not.
        args = default_arg_parser().parse_args()
        if args.debug:
            req_ipdb = PipRequirement('ipdb')
            if not req_ipdb.is_installed():
                logging.error('--debug flag requires ipdb. '
                              'You can install it with:\n%s',
                              ' '.join(req_ipdb.install_command()))
                sys.exit(13)

        if debug or args.debug:
            args.log_level = 'DEBUG'

        # Defer imports so if e.g. --help is called they won't be run
        from coalib.coala_modes import (
            mode_format, mode_json, mode_non_interactive, mode_normal)
        from coalib.output.ConsoleInteraction import (
            show_bears, show_language_bears_capabilities)

        console_printer = ConsolePrinter(print_colored=not args.no_color)
        configure_logging(not args.no_color)

        if args.show_bears:
            from coalib.settings.ConfigurationGathering import (
                get_all_bears)
            filtered_bears = get_all_bears()
            if args.filter_by_language:
                logging.warning(
                    "'--filter-by-language ...' is deprecated. "
                    "Use '--filter-by language ...' instead.")
                if args.filter_by is None:
                    args.filter_by = []
                args.filter_by.append(['language'] + args.filter_by_language)
            if args.filter_by:
                # Each iteration of the following loop applies
                # filters one by one provided as arguments
                try:
                    filtered_bears = apply_filters(
                        args.filter_by, filtered_bears)
                except InvalidFilterException as ex:
                    # If filter is not available
                    console_printer.print(ex)
                    return 2

            local_bears, global_bears = filtered_bears
            show_bears(local_bears,
                       global_bears,
                       args.show_description or args.show_details,
                       args.show_details,
                       console_printer,
                       args)

            return 0
        elif args.show_capabilities:
            from coalib.collecting.Collectors import (
                filter_capabilities_by_languages)
            local_bears, _ = apply_filter('language', args.show_capabilities)
            capabilities = filter_capabilities_by_languages(
                local_bears, args.show_capabilities)
            show_language_bears_capabilities(capabilities, console_printer)

            return 0

        if args.json:
            return mode_json(args, debug=debug)

    except BaseException as exception:  # pylint: disable=broad-except
        if not isinstance(exception, SystemExit):
            if args and args.debug:
                import ipdb
                with ipdb.launch_ipdb_on_exception():
                    raise

            if debug:
                raise

        return get_exitcode(exception)

    if args.format:
        return mode_format(args, debug=debug)

    if args.non_interactive:
        return mode_non_interactive(console_printer, args, debug=debug)

    return mode_normal(console_printer, None, args, debug=debug)


if __name__ == '__main__':  # pragma: no cover
    main()



from coalib.bearlib.aspects import Root


@Root.subaspect
class Security:
    """
    This aspects checks for code with flaws (or security weaknesses) in your
    codebase.
    """
    class docs:
        example = """
        char buf[1024];
        ssizet_t len;
        if ((len = readlink("/modules/pass1", buf, sizeof(buf)-1)) != -1)
            buf[len] = '\0';
        """
        example_language = 'C'
        importance_reason = """
        Security weaknesses can enable malicious users to bypass access controls
        in order to obtain unauthorized privileges, which may result in:
            * Data loss or corruption
            * Denial of access
            * Complete host system takeover
        """
        fix_suggestions = """
        Some few ways to reduce the chance of a vulnerability being used
        against a system are: the use methods that check on availability of
        memory prior to writing, the eradication of null field access and null
        method call, the eradication of redundant condition (condition whose
        result depends on which part is executed first) etc...
        """



from coalib.results.Result import Result
from coalib.results.result_actions.ResultAction import ResultAction

from coala_utils.decorators import enforce_signature


class PrintAspectAction(ResultAction):

    @staticmethod
    @enforce_signature
    def is_applicable(result: Result, original_file_dict, file_diff_dict):
        if result.aspect is None:
            return 'There is no aspect associated with the result.'
        return True

    def apply(self, result, original_file_dict, file_diff_dict):
        """
        Print Aspec(T) Information
        """
        print(type(result.aspect).__qualname__ + '\n' +
              type(result.aspect).docs.definition)

        return file_diff_dict



from coalib.core.Bear import Bear
from coalib.settings.FunctionMetadata import FunctionMetadata


class ProjectBear(Bear):
    """
    This bear base class does not parallelize tasks at all, it runs on the
    whole file base provided.
    """

    def __init__(self, section, file_dict):
        """
        :param section:
            The section object where bear settings are contained. A section
            passed here is considered to be immutable.
        :param file_dict:
            A dictionary containing filenames to process as keys and their
            contents (line-split with trailing return characters) as values.
        """
        Bear.__init__(self, section, file_dict)

        self._kwargs = self.get_metadata().create_params_from_section(section)

    def execute_task(self, args, kwargs):
        # To optimize performance a bit and memory usage, we use args and
        # kwargs from this class instance, instead of passing them via the
        # task.
        return Bear.execute_task(self, (self.file_dict,), self._kwargs)

    @classmethod
    def get_metadata(cls):
        """
        :return:
            Metadata for the ``analyze`` function extracted from its signature.
            Excludes parameters ``self`` and ``files``.
        """
        return FunctionMetadata.from_function(
            cls.analyze,
            omit={'self', 'files'})

    def generate_tasks(self):
        return (tuple(), {}),



"""
Replacement for ``multiprocessing`` library in coala's debug mode.
"""

import logging
import sys
import queue
from functools import partial

from coalib.processes.communication.LogMessage import LogMessage

__all__ = ['Manager', 'Process', 'Queue']


class Manager:
    """
    A debug replacement for ``multiprocessing.Manager``, just offering
    ``builtins.dict`` as ``.dict`` member.
    """

    def __init__(self):
        """
        Just add ``dict`` as instance member.
        """
        self.dict = dict


class Process(partial):
    """
    A debug replacement for ``multiprocessing.Process``, running the callable
    target without any process parallelization or threading.
    """

    def __new__(cls, target, kwargs):
        """
        Just pass the arguments to underlying ``functools.partial``.
        """
        return partial.__new__(cls, target, **kwargs)

    def start(self):
        """
        Just call the underlying ``functools.partial`` instaed of any thread
        or parallel process creation.
        """
        return self()


class Queue(queue.Queue):
    """
    A debug replacement for ``multiprocessing.Queue``, directly processing
    any incoming :class:`coalib.processes.communication.LogMessage.LogMessage`
    instances (if the queue was instantiated from a function with a local
    ``log_printer``).
    """

    def __init__(self):
        """
        Gets local ``log_printer`` from function that created this instance.
        """
        super().__init__()
        # same kind of HACK as can be found in collections.namedtuple for
        # setting .__module__ of created classes
        self.log_printer = sys._getframe(1).f_locals.get('log_printer')

    def put(self, item):
        """
        Add `item` to queue.

        Except `item` is an instance of
        :class:`coalib.processes.communication.LogMessage.LogMessage` and
        there is a ``self.log_printer``. Then `item` is just sent to logger
        instead.
        """
        if isinstance(item, LogMessage):
            logging.log(item.log_level, item.message)
        else:
            super().put(item)



from coalib.core.Bear import Bear
from coalib.settings.FunctionMetadata import FunctionMetadata


class FileBear(Bear):
    """
    This bear base class parallelizes tasks for each file given.
    """

    def __init__(self, section, file_dict):
        """
        :param section:
            The section object where bear settings are contained. A section
            passed here is considered to be immutable.
        :param file_dict:
            A dictionary containing filenames to process as keys and their
            contents (line-split with trailing return characters) as values.
        """
        Bear.__init__(self, section, file_dict)

        self._kwargs = self.get_metadata().create_params_from_section(section)

    def execute_task(self, args, kwargs):
        # To optimize performance and memory usage, we use kwargs from this
        # class instance instead of passing the same settings thousand times in
        # the tasks themselves.
        return Bear.execute_task(self, args, self._kwargs)

    @classmethod
    def get_metadata(cls):
        """
        :return:
            Metadata for the ``analyze`` function extracted from its signature.
            Excludes parameters ``self``, ``filename`` and ``file``.
        """
        return FunctionMetadata.from_function(
            cls.analyze,
            omit={'self', 'filename', 'file'})

    def generate_tasks(self):
        return (((filename, file), {})
                for filename, file in self.file_dict.items())



class CircularDependencyError(RuntimeError):
    """
    An error identifying a circular dependency.
    """

    def __init__(self, names=None):
        """
        Creates the CircularDependencyError with a helpful message about the
        dependency.

        :param names:
            The names of the nodes that form a dependency circle.
        """
        if names:
            msg = 'Circular dependency detected: {names}'.format(
                names=' -> '.join(names))
        else:
            msg = 'Circular dependency detected.'
        super().__init__(msg)
